# -*- coding: utf-8 -*-
# no-pylint: disable=missing-docstring,too-few-public-methods

from pgbouncemgr.logger import format_ex
from pgbouncemgr.postgres import PgReplicationConnection, PgException
from pgbouncemgr.constants import NODE_OFFLINE

class NodePoller():
    """The NodePoller is used to poll all the nodes that are available
       in the state object, and to update their status according to
       the results.
       The connection_class that can be provided is used for testing
       purposes (dependency injection)."""
    def __init__(self, state, log, connection_class=None):
        self.state = state
        self.log = log
        self.connection_class = PgReplicationConnection
        if connection_class is not None:
            self.connection_class = connection_class
        self._connections = {}

    def poll(self):
        """Check the status of all backend nodes and update their status."""
        for node in self.state.nodes.values():
            self._poll_node(node)

    def _poll_node(self, node):
        connection = self._get_connection(node)
        try:
            status = connection.get_replication_status()
            node.status = status["status"]
            if node.status != NODE_OFFLINE:
                node.timeline_id = status["timeline_id"]
                node.system_id = status["system_id"]
        except PgException as exception:
            node.status = NODE_OFFLINE
            self.log.error(
                "[%s] Error on polling node: %s" %
                (node.node_id, format_ex(exception)))

    def _get_connection(self, node):
        if node.node_id in self._connections:
            return self._connections[node.node_id]
        connection = self.connection_class(node.config, self.log)
        self._connections[node.node_id] = connection
        return connection

    def reset(self):
        """Reset all database connections."""
        for connection in self._connections:
            connection.disconnect()
        self._connections = {}
