# -*- coding: utf-8 -*-
# no-pylint: disable=missing-docstring,no-self-use,broad-except

"""This module provides various classes that uare used for connecting to
   a PostgreSQL or pgbouncer server."""

import multiprocessing
import psycopg2
from psycopg2.extras import LogicalReplicationConnection
from pgbouncemgr.logger import format_ex
from pgbouncemgr.constants import \
    NODE_OFFLINE, NODE_STANDBY, NODE_PRIMARY,\
    CONN_REUSED, CONN_RECONNECTED, CONN_CONNECTED


class PgException(Exception):
    """Used for all exceptions that are raised from pgbouncemgr.postgres."""

class PgConnectionFailed(PgException):
    """Raised when connecting to the database server failed."""
    def __init__(self, exception):
        super().__init__(
            "Could not connect to node: %s" % format_ex(exception))

class RetrievingPgReplicationStatusFailed(PgException):
    """Raised when the replication status cannot be determined."""

class ReloadingPgbouncerFailed(PgException):
    """Raised when reloading the pgbouncer configuration fails."""

class ConnectedToWrongBackend(PgException):
    """Raised when the pgbouncer instance is not connected to the
       correct PostgreSQL backend service."""
    def __init__(self, msg):
        super().__init__(
            "The pgbouncer is not connected to the expected PostgreSQL " +
            "backend service: %s" % msg)


# The properties that can be used in a configuration object to
# define connection parameters for psycopg2.
CONNECTION_PARAMS = [
    'host',
    'port',
    'user',
    'password',
    'database',
    'connect_timeout'
]


class NodeLogger():
    def __init__(self, node_config, log):
        self.node_id = node_config.node_id
        self.log = log

    def debug(self, msg):
        self.log.debug("[%s] %s" % (self.node_id, msg))

    def info(self, msg):
        self.log.info("[%s] %s" % (self.node_id, msg))

    def warning(self, msg):
        self.log.warning("[%s] %s" % (self.node_id, msg))

    def error(self, msg):
        self.log.error("[%s] %s" % (self.node_id, msg))


class PgConnection():
    """Implements a connection to a PostgreSQL server.
       The provided node_config must be a pgbouncemgr.NodeConfig object.
       The psycopg2_module argument is used as a depencency injection
       mechanism for testing purposes."""
    def __init__(self, node_config, log, psycopg2_module=psycopg2):
        self.log = NodeLogger(node_config, log)
        self.psycopg2 = psycopg2_module

        self.node_id = node_config.node_id
        self.conn_params = self._create_conn_params(node_config)
        self.ping_query = "SELECT 1"
        self.conn = None
        self.reconnecting = False

    def _create_conn_params(self, node_config):
        """Extract all connection parameters from the provided configuration,
           that don't have value of None."""
        conn_params = {}
        for key in CONNECTION_PARAMS:
            if not hasattr(node_config, key):
                continue
            value = getattr(node_config, key)
            if value is None:
                continue
            conn_params[key] = value
        return conn_params

    def connect(self):
        """Connect to the database server. When a connection exists, then
           check if it is still operational. If yes, then reuse this
           connection. If no, or when no connection exists, then setup a new
           connection.
           Raises an exeption when the database connection cannot be setup.
           returns CONN_CONNECTED, CONN_REUSED or CONN_RECONNECTED when
           the connection was setup successfully."""
        if self.conn is not None:
            try:
                with self.conn.cursor() as cursor:
                    cursor.execute(self.ping_query)
                self.log.debug("reusing connection")
                return CONN_REUSED
            except psycopg2.OperationalError as exception:
                self.log.warning(
                    "connection went away, reconnecting (got exception: %s)" %
                    format_ex(exception))
                self.reconnecting = True
                self.disconnect()
        try:
            self.conn = self.psycopg2.connect(**self.conn_params)
            if self.reconnecting:
                self.log.info("reconnected successfully")
                self.reconnecting = False
                return CONN_RECONNECTED
            self.log.info("connected successfully")

            return CONN_CONNECTED
        except psycopg2.OperationalError as exception:
            self.disconnect()
            self.log.error("connection failed: %s" % format_ex(exception))
            raise PgConnectionFailed(exception)

    def disconnect(self):
        """Disconnect from the database server."""
        try:
            if self.conn is not None:
                self.conn.close()
        except Exception:
            pass
        self.conn = None


class PgReplicationConnection(PgConnection):
    """This PostgresQL connection class is used to setup a replication
       connection to a PostgreSQL database server, which can be used
       to retrieve the replication status for the server."""
    def __init__(self, node_config, log, psycopg2_module=psycopg2):
        super().__init__(node_config, log, psycopg2_module)
        self.conn_params["connection_factory"] = LogicalReplicationConnection

    def get_replication_status(self):
        """Returns the replication status for a node. This is an array,
           containing the keys "status" (NODE_OFFLINE, NODE_PRIMARY or
           NODE_STANDBY), "system_id" and "timeline_id".
           Note: before calling this method, first setup the connection using
           the connect() method."""
        status = {
            "status": None,
            "system_id": None,
            "timeline_id": None
        }

        # Try to connect to the node. If this fails, the status is
        # set to NODE_OFFLINE.
        try:
            self.connect()
        except PgConnectionFailed:
            status["status"] = NODE_OFFLINE
            return status

        # Check if the node is running in primary or standby mode.
        try:
            with self.conn.cursor() as cursor:
                cursor.execute("SELECT pg_is_in_recovery()")
                in_recovery = cursor.fetchone()[0]
            status["status"] = NODE_STANDBY if in_recovery else NODE_PRIMARY
        except psycopg2.OperationalError as exception:
            self.disconnect()
            raise RetrievingPgReplicationStatusFailed(
                "SELECT pg_is_in_recovery() failed: %s" % format_ex(exception))

        # Retrieve system_id and timeline_id.
        try:
            with self.conn.cursor() as cursor:
                cursor.execute("IDENTIFY_SYSTEM")
                row = cursor.fetchone()
            system_id, timeline_id, *_ = row
            status["system_id"] = system_id
            status["timeline_id"] = timeline_id
        except psycopg2.OperationalError as exception:
            self.disconnect()
            self.log.error(
                "retrieving replication status failed: %s" %
                format_ex(exception))
            raise RetrievingPgReplicationStatusFailed(
                "IDENTIFY_SYSTEM failed: %s" % format_ex(exception))

        self.log.debug(
            "node status: status=%s, system_id=%s, timeline_id=%s" %
            (status["status"], status["system_id"], status["timeline_id"]))
        return status

# The query that is used to check if pgbouncer is connected to the expected
# backend server. We don't really have a unique identifier for the node to
# work with there, but we can request the server host + port to see if they
# match those of the expected node. Additionally, a check is done to see if
# the node is running as a primary node.
#
# It would be really nice if it were possible to retrieve the system_id
# and timeline_id for the backend server, but that is only possible when
# using a replication connection. Pgbouncer (through which we are connected
# when running this query) does not support pooled replication connections.
VERIFY_QUERY = """
    SELECT CASE
      WHEN (
        inet_server_addr() IS DISTINCT FROM %(host)s OR
        inet_server_port() IS DISTINCT FROM %(port)s)
      THEN format(
        'connected to wrong server: %%s:%%s (expected %%s:%%s)',
        inet_server_addr(), inet_server_port(), %(host)s, %(port)s)
      WHEN pg_is_in_recovery()
      THEN format(
        'server %%s:%%s is not running as primary',
        %(host)s, %(port)s)
      ELSE NULL
    END
"""

class PgConnectionViaPgbouncer(PgConnection):
    """This PostgreSQL connection class is used to setup a connection
       to the PostgreSQL cluster, via the pgbouncer instance."""
    def __init__(self, node_config, pgbouncer_config, log, psycopg2_module=psycopg2):
        """Instantiate a new connection. The node_config and the
           pgbouncer_config will be combined to get the connection parameters
           for connecting to the PostgreSQL server."""
        self.node_config = node_config

        # First, apply all the connection parameters as defined for the node.
        # This is fully handled by the parent class.
        super().__init__(node_config, log, psycopg2_module)

        # Secondly, override parameters to redirect the connection to
        # the pgbouncer instance.
        self.conn_params["host"] = pgbouncer_config["host"]
        self.conn_params["port"] = pgbouncer_config["port"]

        # Note that we don't setup a replication connection here. This is
        # unfortunately not possible, because pgbouncer does not support
        # this type of connection. If this would ever become possible in
        # pgbouncer, I will definitely switch to such connection, since it
        # allows for doing some extra checks.

    def verify_connection(self):
        """Check if the connection via pgbouncer ends up with the
           configured node."""
        # This is done in a somewhat convoluted way with a subprocess and a
        # timer. This is done, because a connection is made through
        # pgbouncer, and pgbouncer will try really hard to connect the user
        # to its known backend service. When that service is unavailable,
        # then pgbouncer accepts the connection, but the communication will
        # then stall for quite a while. For swift operation of backend
        # switching, we therefore use the timeout setup here, to reconfigure
        # the system more quickly.
        #
        # Note that in most situations this shouldn't be an issue, since we
        # will always reload pgbouncer after a configuration change and the
        # connection from below will, because of that, work as intended
        # right away. We must be prepared for the odd case out though, since
        # we're going for HA here."""
        def check_func(report_func):
            # Setup the database connection
            try:
                self.connect()
            except Exception as exception:
                return report_func(False, exception)

            # Check if we're connected to the requested node.
            with self.conn.cursor() as cursor:
                try:
                    cursor.execute(VERIFY_QUERY, {
                        "host": self.node_config.host,
                        "port": self.node_config.port
                    })
                    result = cursor.fetchone()[0]
                    self.disconnect()

                    if result is not None:
                        raise ConnectedToWrongBackend(result)
                except Exception as exception:
                    self.disconnect()
                    return report_func(False, exception)

            # When the verify query did not return an error message, then we
            # are in the green.
            return report_func(True, None)

        parent_conn, child_conn = multiprocessing.Pipe()
        def report_func(true_or_false, exception):
            child_conn.send([true_or_false, exception])
            child_conn.close()
        proc = multiprocessing.Process(target=check_func, args=(report_func,))
        proc.start()
        proc.join(self.node_config.connect_timeout)
        if proc.is_alive():
            proc.terminate()
            proc.join()
            return (False, PgConnectionFailed("Connection attempt timed out"))
        result = parent_conn.recv()
        proc.join()
        ##### DEBUG
        raise Exception(repr(result))
        ##### /DEBUG
        return result


class PgBouncerConsoleConnection(PgConnection):
    """This PostgreSQL connection class is used to setup a console
       connection to a pgbouncer server. This kind of connection can be
       used to control the pgbouncer instance via admin commands.
       This connection is used by pgbouncemgr to reload the configuration
       of pgbouncer when the cluster state changes."""
    def __init__(self, node_config, log, psycopg2_module=psycopg2):
        super().__init__(node_config, log, psycopg2_module)

        # For the console connection, the database name "pgbouncer"
        # must be used.
        self.conn_params["database"] = "pgbouncer"

        # The default ping query does not work when connected to the
        # pgbouncer console. Here's a simple replacement for it.
        self.ping_query = "SHOW VERSION"

    def connect(self):
        """Connect to the pgbouncer console. Enable the autocommit feature on
           the connection, so the  underlying PostgreSQL client library won't
           automatically try to setup a transaction. Transactions are not
           supported by the pgbouncer console."""
        result = super().connect()
        self.conn.autocommit = True
        return result

    def reload(self):
        """Send the 'RELOAD' command to the pgbouncer console, in order to
           reload the configuration file."""
        try:
            self.connect()
            with self.conn.cursor() as cursor:
                cursor.execute('RELOAD')
                if cursor.statusmessage == 'RELOAD':
                    return
                raise ReloadingPgbouncerFailed(
                    "Unexpected status message: %s" % cursor.statusmessage)
        except Exception as exception:
            raise ReloadingPgbouncerFailed(
                "An exception occurred: %s" % format_ex(exception))
